"""
ABCs for Transits and Traffic.
"""
import abc

class Transit(metaclass = abc.ABCMeta):
	"""
	The abstract Transit.

	Transits are objects that facilitate arbitrary transfers. In the general form, the
	objects or data transferred is unknown; specific subclasses ultimately determine the
	type of data or objects that are transferred.
	"""
	@abc.abstractproperty
	def resource(self):
		"""
		The resource currently acquired by the Transit instance.
		"""

	@abc.abstractproperty
	def traffic(self):
		"""
		The Traffic instance that the Transit is participating in.

		:py:obj:`None` if the Transit has not been attached to a Traffic instance.
		"""

	@abc.abstractproperty
	def polarity(self):
		"""
		Whether the Transit gains or loses during transfers. (receives versus sends)

		:py:obj:`traffic.protocol.polarity.gain` or
		:py:obj:`traffic.protocol.polarity.lose` or
		:py:obj:`traffic.protocol.polarity.none`.
		"""

	@abc.abstractmethod
	def acquire(self, resource):
		"""
		:param resource: The resource to be acquired. The Transit subclass dictates the required type.
		:type resource: :py:class:`object`

		Acquire a resource for facilitating Transfers.
		"""

	@abc.abstractmethod
	def transfer(self):
		"""
		:returns: Description of the transfer that happened during the current cycle.
		:rtype: :py:class:`object`

		Get the transfer information from the transit. Usually, this method should only be
		called during an open Traffic cycles.
		"""

	@abc.abstractmethod
	def terminate(self):
		"""
		Terminate the Transit ultimately signalling the cessation of Transfer Events.
		"""

class Traffic(metaclass = abc.ABCMeta):
	"""
	The abstract Traffic.
	"""
	@abc.abstractproperty
	def polarity(self):
		"""
		Always :py:obj:`traffic.protocol.polarity.none`.

		The Traffic instance only associates itself with the Transits performing Transfers;
		The individual Transits have normal polarity.
		"""

	@abc.abstractmethod
	def acquire(self, transit):
		"""
		:param transit: The :py:class:`Transit` object to be acquired.
		:type transit: :py:class:`Transit`

		Traffic resources are Transits and implicitly acquires itself.
		"""

	@abc.abstractproperty
	def volume(self):
		"""
		The number of Transits participating in this Traffic instance.

		This attribute is never less than ``1``.
		"""

	@abc.abstractmethod
	def force(self):
		"""
		In cases where the Traffic instance is waiting for events, this method provides a
		mechanism to force the Traffic instance to yield the transfers that have accumulated
		so far. Normally, this will result in an empty Traffic cycle.
		"""
